#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest  # noqa

from qary.clibot import CLIBot

__author__ = "SEE AUTHORS.md"
__copyright__ = "Hobson Lane"
__license__ = "The Hippocratic License, see LICENSE.txt (MIT + Do no Harm)"


def test_clibot():
    clibot = CLIBot()
    assert callable(clibot.reply)
    assert isinstance(clibot.reply('Hi'), str)



""" Search and scrape wikipedia articles from a chosen category

This is how secure access to es works using basic auth

```python
from elasticsearch import Elasticsearch, NotFoundError
ES_HOST = 'es.qary.ai'
ES_PORT = 80
ES_USER = 'kibadmin'
ES_PASS = 'p@ss'
client = Elasticsearch(f"http://{ES_USER}:{ES_PASS}@{ES_HOST}:{ES_PORT}")
client.indices.exists('wikipedia')
```

"""
import os
import requests
from pathlib import Path

from elasticsearch import Elasticsearch, NotFoundError
import wikipediaapi
from slugify import slugify
import numpy as np
import pandas as pd
import yake

from qary import constants
from qary.constants import ES_URL, ES_INDEX, ARTICLES_URL

log = constants.logging.getLogger(__name__)


def connect(es_url=ES_URL):
    client = None
    if es_url:
        try:
            client = Elasticsearch(es_url)
        except ConnectionRefusedError:
            log.error("Failed to connect ElasticSearch client")
    return client


client = connect()


wiki_wiki = wikipediaapi.Wikipedia('en')
kw_extractor = yake.KeywordExtractor()


def create_index(index=ES_INDEX):
    if client and not client.indices.exists(index=index):
        log.info(f"No index {index} found, creating it ...")
        client.indices.create(index=ES_INDEX, ignore=400)
    if client.indices.exists(index=index):
        return index
    log.warning(f"Unable to create index {index} using {client}")


def extract_keywords(text):
    keywords = kw_extractor.extract_keywords(text)
    kw = ','.join([kw[1] for kw in keywords])
    return kw


def load_parsed_articles_from_url(url=ARTICLES_URL, filename='parsed_wiki_articles.pkl'):
    r = requests.get(url)
    open(filename, 'wb').write(r.content)


def get_articles_from_pkl(filedir='parsed_wiki_articles.pkl', index=ES_INDEX):
    unpickled_df = pd.read_pickle(filedir)
    articles = unpickled_df.to_dict('records')
    return articles


def add_parsed_document_to_elasticsearch(sections, index=ES_INDEX):
    pageid = sections[0]['page_id']
    if not client:
        return
    exists = None
    try:
        exists = in_index(article_id=pageid, index=index)
    except NotFoundError as e:
        log.warning(f'It looks like the {index} does not exist (NotFoundError):')
        log.error(e)

    if not exists:
        log.info(f"Loading article with page id {pageid} into ElasticSearch index {index}...")
        for section in sections:
            try:
                client.index(index=index, body=section)
            except Exception as err:
                log.error(f'Failed to add the article with page ID {pageid} to {index}', err)
        return sections
    else:
        log.info(f'Article with page_id {pageid} is already in the index.')


def print_categorymembers(categorymembers, level=0, max_level=1):
    for c in categorymembers.values():
        print("%s: %s (ns: %d)" % ("*" * (level + 1), c.title, c.ns))
        if c.ns == wikipediaapi.Namespace.CATEGORY and level < max_level:
            print_categorymembers(c.categorymembers, level=level + 1, max_level=max_level)


def save_articles(
        path=os.path.join(constants.DATA_DIR, "wikipedia"),
        category='Natural_language_processing'):
    """ Save articles in separate .txt files """
    os.makedirs(path, exist_ok=True)
    wiki_wiki = wikipediaapi.Wikipedia('en')
    cat = wiki_wiki.page(f"Category:{category}")

    for key, value in cat.categorymembers.items():
        page = wiki_wiki.page(key)
        slug = slugify(key)
        try:
            with open(f'{path}/{slug}.txt', 'w') as new_file:
                new_file.write(page.title)
                new_file.write('\n')
                new_file.write(page.fullurl)
                new_file.write('\n')
                new_file.write(page.text)
        except Exception as error:
            log.error(f"Error writing document {page.title}: {error}")


def index_dir(path=Path(constants.DATA_DIR, 'corpora', 'wikipedia')):
    """ NotImplemented... yet"""
    raise NotImplementedError(
        "Walks a directory tree and returns a list of paths but doesn't do anything with them.")
    paths = []
    # r=root, d=directories, f = files
    for r, d, f in os.walk(path):
        for file in f:
            if '.txt' in file:
                paths.append(os.path.join(r, file))

    return paths


class Document:

    def __init__(self, title, text, source):

        self.title = title
        self.text = text
        self.source = source
        self.body = {}
        try:
            self.body = {
                "title": self.title,
                "text": self.text,
                "source": self.source
            }
            log.info(f"Elasticsearch Document JSON created: {self.body}")
        except Exception as error:
            log.warning(f"Document JSON instance error: ", error)

    def insert(self):
        slug = slugify(self.body['title'])
        if client:
            try:
                return client.index(index=slug, body=self.body)
            except Exception as e:
                log.error(f"Could not create a JSON entry for an article {slug}. error: {e}")


def search(text=None, title=None, index=''):
    """ Full text search within an ElasticSearch index (''=all indexes) for the indicated text """
    if not client:
        return {}
    if title:
        body = {"query": {"match": {"title": title}}}
    else:
        body = {"query": {"match": {"text": text}}}
    try:
        results = client.search(index=index, body=body)
    except NotFoundError:
        return {}
    return results


def in_index(article_id=None, title=None, index=ES_INDEX):
    """ Check if the article is already in the ElasticSearch index by looking it up by page_id

    >>> bool(in_index(article_id=123, index=''))
    False
    >>> x = in_index(article_id=31491224, index='')
    >>> (x is True) or (x is None)
    True
    >>> x = in_index(title='Trevor Noah', index='')
    >>> (x is True) or (x is None)
    True
    """
    if not client:
        return
    if title:
        body = {"query": {"match": {"article_title": title}}}
    else:
        body = {"query": {"match": {"page_id": article_id}}}
    try:
        results = client.search(index=index, body=body)
    except NotFoundError:
        return False
    return len(results.get('hits', {}).get('hits', {})) > 0


def deconstruct_wikipedia_article(page=None, pageid=None, title=None):
    page = page or wiki_wiki.page(pageid or title)
    if not in_index(page.pageid):
        sections = deconstruct_article(page=page)
        return sections


def deconstruct_article(page=None, title=None):
    """ This is horribly recursive """
    page = page or wiki_wiki.page(title)
    section_list = []
    section_list = [{'level': 0,
                     'section_title': 'Summary',
                     'text': page.summary}]

    def get_sections(sections, level=0):
        for s in sections:
            section_dict = {'level': level,
                            'section_title': s.title,
                            'text': s.text}
            section_list.append(section_dict)
            get_sections(s.sections, level + 1)

    get_sections(page.sections)
    return section_list


def build_elasticsearch_record(page, section_list):

    # Transform list of dictionaries to dataframe
    df = pd.DataFrame(section_list)

    # Create column "main_section"
    df['main_section'] = np.nan
    df.loc[df['level'] == 0, 'main_section'] = df['section_title']
    df['main_section'].fillna(method='ffill', inplace=True)

    # Create column "subsection"
    df['subsection'] = np.nan
    df.loc[df['text'] == '', 'subsection'] = df['section_title']
    df['subsection'].fillna(method='ffill', inplace=True)

    # Add wikipedia article title, source url and page id
    df1 = df.fillna('')
    df1['article_title'] = page.title
    df1['source_url'] = page.fullurl
    df1['page_id'] = page.pageid

    # Create a list of section tags
    df1['tags'] = df1.apply(lambda row: [row['article_title'],
                                         row['main_section'],
                                         row['subsection'],
                                         row['section_title']],
                            axis=1)
    df1['tags'] = df1['tags'].apply(lambda cell: [s for s in cell if s != ""])
    df1['tags'] = df1['tags'].apply(lambda cell: list(dict.fromkeys(cell)))
    df1['keywords'] = df1.apply(lambda x: extract_keywords(x['text']) if len(x['text']) > 1000 else "", axis=1)

    # Drop rows with NaN values (empty sections)
    df2 = df1.replace('', np.nan, regex=True)
    df2 = df2.drop(['level', 'subsection'], axis=1).dropna()

    # Transform a list of tags to a comma separated string
    df2['tags'] = df2.apply(lambda row: ','.join(row['tags']), axis=1)

    # Add tags as the first line of the text field
    df2['text'] = df2.apply(lambda row: row['tags'] + '\n' + row['text'], axis=1)

    # Add number of section withing the article
    df2['section_number'] = df2.index

    return df2.to_dict(orient='records')


def parse_and_index_wikipedia_article(title='Trevor Noah', index=ES_INDEX):
    page = wiki_wiki.page(title)
    sections = deconstruct_wikipedia_article(page=page) or []
    records = build_elasticsearch_record(page, sections) or []
    return add_parsed_document_to_elasticsearch(sections=records, index=index)



import os
from pathlib import Path

from tqdm import tqdm
import yaml
from yaml.scanner import ScannerError

import numpy as np

from qary.constants import FAQ_DOMAINS, DATA_DIR
from qary.spacy_language_model import nlp, UNKNOWN_WORDVEC


import logging
log = logging.getLogger(locals().get('__name__'))


def normalize_docvectors(docvectors):
    """ Convert a table (2D matrix) of row-vectors into a table of normalized row-vectors

    >>> vecs = normalize_docvectors([[1, 2, 3], [4, 5, 6], [0, 0, 0], [-1, 0, +2]])
    >>> vecs.shape
    (4, 3)
    >>> np.linalg.norm(vecs, axis=1).round()
    array([1., 1., 0., 1.])
    """
    docvectors = np.array(docvectors)
    log.info(f'docvectors.shape: {docvectors.shape}')
    norms = np.linalg.norm(docvectors, axis=1)
    iszero = norms <= 0
    log.info(f'norms.shape: {norms.shape}')
    norms_reshaped = norms.reshape(-1, 1).dot(np.ones((1, docvectors.shape[1])))
    log.info(f'norms_reshaped.shape: {norms_reshaped.shape}')
    if np.any(iszero):
        log.info(
            f'Some doc vectors are zero like this first one: docvectors[{iszero},:] = {docvectors[iszero,:]}')
    norms_reshaped[iszero, :] = 1
    normalized_docvectors = docvectors / norms_reshaped
    log.info(f'normalized_docvectors.shape: {normalized_docvectors.shape}')
    assert normalized_docvectors.shape == docvectors.shape
    return normalized_docvectors


def load(domains=FAQ_DOMAINS):
    """ Load yaml file, use hashtags to create context tags as multihot columns

    Load faq*.yml into dictionary: question: answer

    >>> g = load()
    >>> len(g['questions']) == len(g['answers']) > 30
    True
    """
    questions, answers, question_vectors = [], [], []

    faqdirpath = os.path.join(DATA_DIR, 'faq')
    for domain in tqdm(domains):
        for filepath in Path(faqdirpath).glob(f'faq*{domain}*.yml'):
            # filename = filepath.name
            try:
                filepointer = open(filepath)
            except FileNotFoundError as e:
                log.error(f"{e}\n    Unable to find the file path object: {filepath}")
                continue
            with filepointer:
                log.info(f"loading: {filepath.name}\n    with file pointer: {filepointer}")
                try:
                    qa_list = yaml.safe_load(filepointer)
                except ScannerError as e:
                    log.error(f"{e}\n    yaml.safe_load unable to read {filepointer.name}")
                    continue

            for i, qa_dict in enumerate(qa_list):
                if qa_dict is None:
                    log.warning(f'Found None instead of dict in FAQ pair {filepointer.name}[{i}].')
                    qa_dict = {}
                questions.append(qa_dict.get('Q', qa_dict.get('q', '')))
                answers.append(qa_dict.get('A', qa_dict.get('a', '')))
                try:
                    question_vectors.append(list(nlp(questions[-1] or '').vector))
                except TypeError:
                    question_vectors.append(list(UNKNOWN_WORDVEC))
                    continue
                assert len(UNKNOWN_WORDVEC) == len(question_vectors[-1])

    log.debug(f'len(question_vectors): {len(question_vectors)}')

    questions = np.array(questions)
    log.debug(f'len(questions): {len(questions)}')
    answers = np.array(answers)
    log.debug(f'len(answers): {len(answers)}')
    mask = np.array([(bool(a) and bool(q) and len(str(a).strip()) > 0 and len(str(q).strip()) > 0)
                     for a, q in zip(questions, answers)])

    question_vectors = normalize_docvectors(question_vectors)

    # This should be a Kendra/gensim/annoy class (with methods like .find_similar)
    return dict(
        questions=questions[mask],
        answers=answers[mask],
        question_vectors=np.array([qv for qv, m in zip(question_vectors, mask) if m])
    )



from collections import Counter, OrderedDict, Mapping
from decimal import Decimal
import datetime
import logging
import os
from pytz import timezone
import requests
import string
import sys

import numpy as np
import pandas as pd

import nltk
import nltk.corpus
import spacy  # noqa
import configargparse
from environment import Environment
import django.conf

from qary import __version__

# DO ACCESS KEY and SECRET need to be integrated into env
env = Environment(spacy_lang=str, loglevel=int, name=str)
# if you need consistent hashes from builtin hash function set env['PYTHONHASHSEED']

HOME_DIR = os.path.expanduser('~')
SRC_DIR = os.path.dirname(__file__)
BASE_DIR = os.path.dirname(SRC_DIR)
DATA_DIR = os.path.join(SRC_DIR, 'data')
LOG_DIR = os.path.join(DATA_DIR, 'log')
HISTORY_PATH = os.path.join(DATA_DIR, 'history.json')
os.makedirs(LOG_DIR, exist_ok=True)
# with open(HISTORY_PATH, 'a') as fap:
#     pass
MIDATA_DOMAINNAME = 'tan.sfo2.digitaloceanspaces.com'
MIDATA_URL = f'https://{MIDATA_DOMAINNAME}'
MIDATA_QA_MODEL_DIR = 'midata/public/models/qa'
MIDATA_QA_MODEL_DIR_URL = f'{MIDATA_URL}{MIDATA_QA_MODEL_DIR}'
ARTICLES_URL = f'{MIDATA_URL}/midata/public/corpora/wikipedia/articles_with_keywords.pkl'

USE_CUDA = False
MAX_TURNS = 10000
EXIT_COMMANDS = set('exit quit bye goodbye cya'.split())


DEFAULT_CONFIG = {
    'name': 'bot',
    'persist': 'False',  # Yes, yes, 1, Y, y, T, t
    'bots': 'glossary',  # glossary,qa,parul,eliza,search_fuzzy'
    'spacy_lang': 'en_core_web_sm',
    'use_cuda': USE_CUDA,
    'loglevel': 0,  # 0 allows the user to set a value 10-50 (lower more verbose)
    'num_top_replies': 10,
    'self_score': '.5',
    'semantic_score': '.5',
    'debug': True,
    'wiki_title_max_words': 4,
    'score_weights': '{"spell": .25, "sentiment": .25, "semantic": .5}',
    'qa_model': 'albert-large-v2-0.2.0',
}
DEFAULT_CONFIG.update(env.parsed)

LOGLEVELS = [
    logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.FATAL]
# LOG_LEVELS = [         10,           20,              30,            40,            50]
LOGLEVEL_NAMES = 'DEBUG INFO WARNING ERROR FATAL'.split()
LOGLEVEL_ABBREVIATIONS = [s[:4].lower() for s in LOGLEVEL_NAMES]
LOGLEVEL_ABBR_DICT = dict(zip(LOGLEVEL_ABBREVIATIONS, LOGLEVELS))
# this is the LOGLEVEL for the top of this file, once args and .ini file are read, it will change
LOGLEVEL = getattr(env, 'loglevel', DEFAULT_CONFIG.get('loglevel', logging.WARNING))
USE_CUDA = getattr(env, 'use_cuda', DEFAULT_CONFIG.get('use_cuda', USE_CUDA))

logging.basicConfig(
    format='%(asctime)s.%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d:%H:%M:%S',
    level=LOGLEVEL)
root_logger = logging.getLogger()
log = logging.getLogger(__name__)

# TZ constants
DEFAULT_TZ = timezone('UTC')

MAX_LEN_FILEPATH = 1023  # on OSX `open(fn)` raises OSError('Filename too long') if len(fn)>=1024

ROUNDABLE_NUMERIC_TYPES = (float, int, Decimal, bool)
FLOATABLE_NUMERIC_TYPES = (float, int, Decimal, bool)
BASIC_NUMERIC_TYPES = (float, int)
NUMERIC_TYPES = (float, int, Decimal, complex, str)  # datetime.datetime, datetime.date
NUMBERS_AND_DATETIMES = (float, int, Decimal, complex, str)
SCALAR_TYPES = (float, int, Decimal, bool, complex, str)  # datetime.datetime, datetime.date
# numpy types are derived from these so no need to include numpy.float64, numpy.int64 etc
DICTABLE_TYPES = (Mapping, tuple, list)  # convertable to a dictionary (inherits Mapping or is a list of key/value pairs)
VECTOR_TYPES = (list, tuple)
PUNC = str(string.punctuation)

# synonyms for "count"
COUNT_NAMES = ['count', 'cnt', 'number', 'num', '#', 'frequency', 'probability', 'prob', 'occurences']
# 4 types of

###########
# secrets
SPACES_ACCESS_KEY = os.environ.get('SPACES_ACCESS_KEY')
SPACES_SECRET_KEY = os.environ.get('SPACES_SECRET_KEY')

ES_HOST = os.environ.get('ES_HOST')
ES_PORT = os.environ.get('ES_PORT', 9200)
ES_USER = os.environ.get('ES_USER', os.environ.get('KIBANA_USER', 'kibadmin'))
ES_PASS = os.environ.get('ES_PASS', os.environ.get('KIBANA_PASS', 'YoUrPaSsWoRd'))
ES_INDEX = os.environ.get('ES_INDEX', 'wikipedia')

# secrets
############


def parse_args(args):
    """Parse command line parameters using qary.ini for the default values

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """

    parser = configargparse.ArgParser(
        default_config_files=[
            '~/qary.ini',
            # '~/qary.ini',
            # '~/nlpiabot.ini',
            # '~/nlpia.ini',
            # os.path.join(BASE_DIR, '*.ini'),
            os.path.join(DATA_DIR, '*.ini'),
        ],
        description="Command line bot application. Try `$ bot how do you work?`")
    parser.add('-c', '--config', required=False, is_config_file=True,
               help="Config file path (default: ~/qary.ini)")
    parser.add_argument(
        '-d', '--debug',
        help="Set DEBUG logging level and raise more exceptions immediately.",
        dest="debug",
        default=str(DEFAULT_CONFIG['debug'])[0].lower() in 'fty1p',
        action='store_true')
    parser.add_argument(
        '--version',
        action='version',
        version='qary {ver}'.format(ver=__version__))
    parser.add_argument(
        '--name',
        default=None,  # DEFAULT_CONFIG['name'],
        dest="nickname",
        help="IRC nick or CLI command name for the bot",
        type=str,
        metavar="STR")
    parser.add_argument(
        '-n',
        '--num_top_replies',
        default=None,  # DEFAULT_CONFIG['num_top_replies'],
        dest="num_top_replies",
        help="Limit on the number of top (high score) replies that are randomly selected from.",
        type=int,
        metavar="INT")
    parser.add_argument(
        '-u',
        '--use_cuda',
        help="Use CUDA and GPU to speed up transformer inference.",
        dest='use_cuda',
        default=USE_CUDA,
        action='store_true')
    parser.add_argument(
        '-p',
        '--persist',
        help="DEPRECATED: Don't exit. Retain language model in memory and maintain dialog until user says 'exit' or 'quit'",
        dest='persist',
        default=str(DEFAULT_CONFIG['persist'])[0].lower() in 'fty1p',
        action='store_true')
    parser.add_argument(
        '-b',
        '--bots',
        default=None,  # DEFAULT_CONFIG['bots'],  # None so config.ini can populate defaults
        dest="bots",
        help="Comma-separated list of bot personalities to load. Defaults: pattern,parul,search_fuzzy,time,eliza",
        type=str,
        metavar="STR")
    parser.add_argument(
        '-s',
        '--skills',
        default=None,  # DEFAULT_CONFIG['bots'],  # None so config.ini can populate defaults
        dest="bots",
        help="Comma-separated list of bot personalities to load. Defaults: pattern,parul,search_fuzzy,time,eliza",
        type=str,
        metavar="STR")
    parser.add_argument(
        '-q',
        '--quiet',
        dest="verbosity",
        help="Quiet: set loglevel to ERROR",
        action='store_const',
        const=logging.ERROR)
    parser.add_argument(
        '-qq',
        '--very_quiet',
        dest="verbosity",
        help="Very quiet: set loglevel to FATAL",
        action='store_const',
        const=logging.FATAL)
    parser.add_argument(
        '-v',
        '--verbose',
        dest="verbosity",
        help="Verbose: set loglevel to INFO",
        action='store_const',
        const=logging.INFO)
    parser.add_argument(
        '-vv',
        '--very_verbose',
        dest="verbosity",
        help="Verty verbose: set loglevel to DEBUG",
        action='store_const',
        const=logging.DEBUG)
    parser.add_argument(
        '-l',
        '--loglevel',
        dest="loglevel",
        help="Raw integer loglevel (10=debug, 20=info, 30=warn, 40=error, 50=fatal)",
        type=int,
        default=0)  # DEFAULT_CONFIG['loglevel'])
    parser.add_argument(
        '--spacy_lang',
        default=None,  # None allows ini to set default
        dest="spacy_lang",
        help="SpaCy language model: en_core_web_sm, en_core_web_md, or en_core_web_lg",
        type=str,
        metavar="STR")
    parser.add_argument(
        '--wiki_title_max_words',
        default=DEFAULT_CONFIG['wiki_title_max_words'],
        dest="wiki_title_max_words",
        help='Maximum n-gram length (in tokens) for wikipedia article title guesses.',
        type=int,
        metavar="INT")
    parser.add_argument(
        '--semantic',
        type=float,
        default=1.0,
        dest='semantic',
        metavar='FLOAT',
        help='set weight of the semantic quality score')
    parser.add_argument(
        '--sentiment',
        type=float,
        default=0.5,
        dest='sentiment',
        metavar='FLOAT',
        help='set weight of the sentiment quality score')
    parser.add_argument(
        '--spell',
        type=float,
        default=0.2,
        dest='spell',
        metavar='FLOAT',
        help='set weight of the spell quality score')
    parser.add_argument(
        'words',
        type=str,
        nargs='*',
        help="Words to pass to bot as an utterance or conversational statement requiring a bot reply or action.")
    parser.add_argument(
        '--qa_model',
        help="Select which model qa_bots will use",
        dest='qa_model',
        default=DEFAULT_CONFIG['qa_model'],
        type=str,
        metavar='STR')
    parsed_args = parser.parse_args(args)
    # print(parsed_args)
    return parsed_args


def setup_logging(loglevel=LOGLEVEL):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    global LOGLEVEL, log, root_logger

    logformat = '%(asctime)s.%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s'
    logdatefmt = "%Y-%m-%d%H:%M:%S"
    root_logger.setLevel(loglevel)
    log.setLevel(loglevel)
    # FIXME: this doesn't seem to change things, the original format and level set at top of file rules
    logging.basicConfig(level=loglevel, stream=sys.stdout, format=logformat, datefmt=logdatefmt)
    # set the root logger to the same log level
    logging.getLogger().setLevel(loglevel)


def parse_argv(argv=sys.argv):
    """ Parse the command line args and ini file. Business logic to resolve conflicting arg values """
    global BOT, USE_CUDA, log

    new_argv = []
    if len(argv) > 1:
        command_line_path = str(argv[0]).lower().strip()
        if command_line_path.endswith('qary') or command_line_path.endswith('bot'):
            new_argv.extend(list(argv[1:]))
    else:
        log.info(f"It doesn't look like you're running this as a command line application. sys.argv: {argv}")
    args = parse_args(args=new_argv)

    # consolidate the 2 synonymous args, loglevel and verbosity, by using the minimum of the 2
    #   loglevel may be set by the ini file or the command line arg loglevel
    #   verbosity may only be set by the command line args -v/verbose -qq -vv and -q/quiet
    #   The more verbose of the 2 (lower loglevel value) wins

    loglevel = min(args.loglevel or 0, args.verbosity or 0) or logging.WARNING
    setup_logging(loglevel=loglevel)
    log.debug(f'RAW ARGS (including config file): {vars(args)}')
    args.loglevel = loglevel

    # strip quotes in case ini file incorrectly uses single quotes that become part of the str
    args.nickname = str(args.nickname).strip().strip('"').strip("'")
    args.bots = 'glossary,qa' if getattr(args, 'bots', getattr(args, 'skills', None)) is None else args.bots
    args.bots = [m.strip() for m in args.bots.split(',')]
    log.debug(f"Building a BOT with: {args.bots}")

    USE_CUDA = args.use_cuda

    return args


try:
    # This will fail if another application (like gunicorn) imports qary and redirects stdin without
    # running it as a command line app
    print()  # can't log anything until we know what the user's log level is.
    # print(f"sys.argv: {sys.argv}")
    args = parse_argv(argv=sys.argv)
except Exception as e:  # noqa
    log.info(e)
    log.info('Unable to parse command line arguments. Are you trying to import this into gunicorn?')
    # Workaround for the bug when Django app tries to import qary.constants:
    # `usage: gunicorn [-h] [-c CONFIG] ... gunicorn: error: unrecognized arguments
    args = parse_argv(argv=[])


LOGLEVEL = args.loglevel or LOGLEVEL


# handler = logging.handlers.TimedRotatingFileHandler(os.path.join(LOG_DIR, 'qary.constants.log'), when='midnight')
# handler.setLevel(logging.INFO)
# log.addHandler(handler)


LANGS = ['en_core_web_sm', 'en_core_web_md', 'en_core_web_lg']
LANGS_ABBREV = 'en enmd enlg'.split()
LANGS += 'de_core_news_sm de_core_news_md de_trf_bertbasecased_lg'.split()
LANGS_ABBREV += 'de demd delg'.split()

# FOREIGN_LANGS_DF = pd.read_csv(os.path.join(DATA_DIR, 'spacy_languages.csv'))
# for i, row in FOREIGN_LANGS_DF.iterrows():
#     match = re.match(r'(\d)+\b', row['Models'])
#     if match:
#         num_models = int(match.groups()[0])
#         LANGS.extend([row['Code']] * num_models)

# tuple('de de de'.split())  # df=pd.read_html('https://spacy.io/usage/models')[0]
LANGS_ABBREV = dict(zip(LANGS_ABBREV, LANGS))

LANG = getattr(args, 'spacy_lang', None) or DEFAULT_CONFIG.get('spacy_lang') or LANGS[0]
log.info(f'LANG=spacy_lang={LANG}')

try:
    STOPWORDS_DICT = Counter(nltk.corpus.stopwords.words('english'))
except LookupError:
    log.info('Downloading NLTK stopwords and punkt corpora')
    nltk.download('stopwords', quiet=True)
    nltk.download('punkt', quiet=True)
    STOPWORDS_DICT = Counter(nltk.corpus.stopwords.words('english'))
STOPWORDS = set(STOPWORDS_DICT)

QUESTIONWORDS = set('who what when were why which how'.split() + ['how come', 'why does', 'can i', 'can you', 'which way'])
QUESTION_STOPWORDS = QUESTIONWORDS | STOPWORDS

ASCII_LOWER = 'abcdefghijklmnopqrstuvwxyz'
ASCII_UPPER = ASCII_LOWER.upper()


# clibot.py
DEFAULT_BOTS = ['pattern']  # 'search_fuzzy', 'parul', 'eliza', 'glossary', 'qa'

# faq_bots.py
FAQ_DOMAINS = (
    'python-data-science',
)
FAQ_MIN_SIMILARITY = 0.88
FAQ_MAX_NUM_REPLIES = 3

# glossary_bots.py
DEFAULT_GLOSSARY_DOMAINS = ('dsdh', )

TFHUB_USE_MODULE_URL = "https://tfhub.dev/google/universal-sentence-encoder-large/3"

# # templates for medical sentences
# SENTENCE_SPEC_PATH = os.path.join(DATA_DIR, 'medical_sentences.json')
# SENTENCE_SPEC = json.load(open(SENTENCE_SPEC_PATH, 'r'))


# Universal Sentence Encoder's TF Hub module for creating USE Embeddings from
USE = None


def set_django_settings(settings=None):
    DATABASES = {
        'sqlite': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
        },
        'postgres': {
            'ENGINE': os.environ.get('SQL_ENGINE', 'django.db.backends.sqlite3'),
            'NAME': os.environ.get('SQL_DATABASE', os.path.join(BASE_DIR, 'db.sqlite3')),
            'USER': os.environ.get('SQL_USER', 'user'),
            'PASSWORD': os.environ.get('SQL_PASSWORD', 'password'),
            'HOST': os.environ.get('SQL_HOST', 'localhost'),
            'PORT': os.environ.get('SQL_PORT', '5432'),
        }
    }
    DATABASES['default'] = DATABASES['sqlite']
    settings = {'DATABASE_' + k: v for (k, v) in DATABASES['default'].items()}
    settings['TIME_ZONE'] = 'America/Los_Angeles'
    try:
        django.conf.settings.configure(**settings)
    except RuntimeError:  # RuntimeError('Settings already configured.')
        pass
    return django.conf.settings


django_settings = set_django_settings()


#####################################################################################
# pugnlp.constants

tld_iana = pd.read_csv(os.path.join(DATA_DIR, 'constants', 'tlds-from-iana.csv'), encoding='utf8')
tld_iana = OrderedDict(sorted(zip((tld.strip().lstrip('.') for tld in tld_iana.domain),
                                  [(sponsor.strip(), -1) for sponsor in tld_iana.sponsor]),
                              key=lambda x: len(x[0]),
                              reverse=True))
# top 20 in Google searches per day
# sorted by longest first so .com matches before .om (Oman)
tld_popular = OrderedDict(sorted([
    ('com', ('Commercial', 4860000000)),
    ('org', ('Noncommercial', 1950000000)),
    ('edu', ('US accredited postsecondary institutions', 1550000000)),
    ('gov', ('United States Government', 1060000000)),
    ('uk', ('United Kingdom', 473000000)),  # noqa
    ('net', ('Network services', 206000000)),
    ('ca', ('Canada', 165000000)),  # noqa
    ('de', ('Germany', 145000000)),  # noqa
    ('jp', ('Japan', 139000000)),  # noqa
    ('fr', ('France', 96700000)),  # noqa
    ('au', ('Australia', 91000000)),  # noqa
    ('us', ('United States', 68300000)),  # noqa
    ('ru', ('Russian Federation', 67900000)),  # noqa
    ('ch', ('Switzerland', 62100000)),  # noqa
    ('it', ('Italy', 55200000)),  # noqa
    ('nl', ('Netherlands', 45700000)),  # noqa
    ('se', ('Sweden', 39000000)),  # noqa
    ('no', ('Norway', 32300000)),  # noqa
    ('es', ('Spain', 31000000)),  # noqa
    ('mil', ('US Military', 28400000)),
    ], key=lambda x: len(x[0]), reverse=True))

uri_schemes_iana = sorted(pd.read_csv(os.path.join(DATA_DIR, 'constants', 'uri-schemes.xhtml.csv'),
                                      index_col=0).index.values,
                          key=lambda x: len(str(x)), reverse=True)
uri_schemes_popular = ['chrome-extension', 'example', 'content', 'bitcoin',
                       'telnet', 'mailto',
                       'https', 'gtalk',
                       'http', 'smtp', 'feed',
                       'udp', 'ftp', 'ssh', 'git', 'apt', 'svn', 'cvs']

# these may not all be the sames isinstance types, depending on the env
FLOAT_TYPES = tuple([t for t in set(np.typeDict.values()) if t.__name__.startswith('float')] + [float])
FLOAT_DTYPES = tuple(set(np.dtype(typ) for typ in FLOAT_TYPES))
INT_TYPES = tuple([t for t in set(np.typeDict.values()) if t.__name__.startswith('int')] + [int])
INT_DTYPES = tuple(set(np.dtype(typ) for typ in INT_TYPES))
NUMERIC_TYPES = tuple(set(list(FLOAT_TYPES) + list(INT_TYPES)))
NUMERIC_DTYPES = tuple(set(np.dtype(typ) for typ in NUMERIC_TYPES))

DATETIME_TYPES = [t for t in set(np.typeDict.values()) if t.__name__.startswith('datetime')]
DATETIME_TYPES.extend([datetime.datetime, pd.Timestamp])
DATETIME_TYPES = tuple(DATETIME_TYPES)

DATE_TYPES = (datetime.datetime, datetime.date)

# matrices can be column or row vectors if they have a single col/row
VECTOR_TYPES = (list, tuple, np.matrix, np.ndarray)
MAPPING_TYPES = (Mapping, pd.Series, pd.DataFrame)

# These are the valid dates for all 3 datetime types in python (and the underelying integer nanoseconds)
INT_MAX = INT64_MAX = 2 ** 63 - 1
INT_MIN = INT64_MIN = - 2 ** 63
UINT_MAX = UINT64_MAX = - 2 ** 64 - 1

INT32_MAX = 2 ** 31 - 1
INT32_MIN = - 2 ** 31
UINT32_MAX = - 2 ** 32 - 1

INT16_MAX = 2 ** 15 - 1
INT16_MIN = - 2 ** 15
UINT16_MAX = - 2 ** 16 - 1

# Pandas timestamps can handle nanoseconds? but python datetimestamps cannot.
MAX_TIMESTAMP = pd.Timestamp('2262-04-11 23:47:16.854775', tz='utc')
MIN_TIMESTAMP = pd.Timestamp(datetime.datetime(1677, 9, 22, 0, 12, 44), tz='utc')
ZERO_TIMESTAMP = pd.Timestamp('1970-01-01 00:00:00', tz='utc')

# to_pydatetime() rounds to microseconds, ignoring 807 nanoseconds available in other MAX TIMESTAMPs
MIN_DATETIME = MIN_TIMESTAMP.to_pydatetime()
MAX_DATETIME = MAX_TIMESTAMP.to_pydatetime()
MIN_DATETIME64 = MIN_TIMESTAMP.to_datetime64()
MAX_DATETIME64 = MAX_TIMESTAMP.to_datetime64()
INF = np.inf
NAN = np.nan
NAT = pd.NaT


# str constants
MAX_CHR = MAX_CHAR = chr(127)
APOSTROPHE_CHARS = "'`’"
# Monkey patch so import from constants if you want this:
string.unprintable = '\x00\x01\x02\x03\x04\x05\x06\x07\x08\x0e\x0f' \
    '\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f\x7f'
ASCII_UNPRINTABLE_CHRS = string.unprintable  # ''.join(chr(i) for i in range(128) if chr(i) not in string.printable)

NULL_VALUES = set(['0', 'None', 'null', "'"] + ['0.' + z for z in ['0' * i for i in range(10)]])
# if datetime's are 'repr'ed before being checked for null values sometime 1899-12-30 will come up
NULL_REPR_VALUES = set(['datetime.datetime(1899, 12, 30)'])
# to allow NULL checks to strip off hour/min/sec from string repr when checking for equality
MAX_NULL_REPR_LEN = max(len(s) for s in NULL_REPR_VALUES)

PERCENT_SYMBOLS = ('percent', 'pct', 'pcnt', 'pt', r'%')
FINANCIAL_WHITESPACE = ('Flat', 'flat', ' ', ',', '"', "'", '\t', '\n', '\r', '$')
FINANCIAL_MAPPING = (('k', '000'), ('M', '000000'))

# pugnlp.constants
#####################################################################################


def check_es_url(es_host=ES_HOST, es_port=ES_PORT, es_user=ES_USER, es_pass=ES_PASS, es_index=ES_INDEX, es_url=None):
    es_url = es_url or f'http://{ES_USER}:{ES_PASS}@{ES_HOST}:{ES_PORT}'
    # TODO, use regex to check to make sure es_url contains an es_host too
    if es_url:
        es_index_url = f'{es_url}/{es_index}' if es_index else es_url
        try:
            response = requests.get(es_index_url, timeout=3)
            log.debug(f'ElasticSearch Server Replied: {response.json()}')
        except requests.ConnectTimeout:
            log.info(f"ConnectTimeout while trying to connect to {es_index_url}...")
            return None
        except requests.ConnectionError:
            log.info(f"Unable to connect to {es_index_url}...")
            return None
        except requests.Timeout:
            log.info(f"Timeout while trying to connect to {es_index_url}...")
            return None
        except requests.exceptions.InvalidURL:
            log.info(f"Invalid URL: {es_index_url}...")
            return None
    return es_url


ES_URL = check_es_url()
ES_HOST = None if ES_URL is None else ES_HOST



import sys
import logging

import numpy as np
import spacy

from .constants import LANG, LANGS


log = logging.getLogger(__name__)
nlp = None
UNKNOWN_WORDVEC = np.array([])


class passthroughSpaCyPipe:
    """ Callable pass-through SpaCy Pipeline Component class (callable) for fallback if spacy_hunspell.spaCyHunSpell fails"""

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, doc):
        log.info(f"This passthroughSpaCyPipe component only logs the token count: {len(doc)}")
        return doc


try:
    import spacy_hunspell
except ImportError:
    log.debug('Failed to import spaCyHunSpell. Substituting with fake . . .')
    spacy_hunspell = None
spaCyHunSpell = spacy_hunspell.spaCyHunSpell if spacy_hunspell else passthroughSpaCyPipe


def add_hunspell_pipe(model):
    try:
        spacy.tokens.Token.get_extension('hunspell_spell')
        log.debug(f'SpaCy Token already has a hunspell Pipe section . . .')
        return model
    except ValueError:
        pass
    if sys.platform == 'linux' or sys.platform == 'linux2':
        hunspell = spaCyHunSpell(model, 'linux')
    elif sys.platform == 'darwin':
        hunspell = spaCyHunSpell(model, 'mac')
    else:  # sys.platform == 'win32':
        try:
            # TODO determine paths for en_US.dic and en_US.aff on windows
            hunspell = spaCyHunSpell(model, ('en_US.dic', 'en_US.aff'))
        except Exception:
            log.debug('Failed to locate en_US.dic and en_US.aff files. Substituting with fake . . .')
            hunspell = passthroughSpaCyPipe()
    try:
        model.add_pipe(hunspell)
    except ValueError:
        log.debug(f'SpaCy parser {model} already has a hunspell Pipe section...')
    return model


def load(lang=None):
    """ Load the specified language model or the small English model, if none specified

    >>> load()  # doctest: +ELLIPSIS
    <spacy.lang.en.English object at ...
    """
    global nlp, UNKNOWN_WORDVEC
    model = None
    log.info(f"Loading SpaCy model...")
    nlp_lang = getattr(nlp, 'lang', '')
    nlp_meta = getattr(nlp, 'meta', {})
    nlp_size = nlp_meta.get('name', '')[-2:]
    if nlp_lang and (not lang or (nlp_lang == lang[:2] and nlp_size == lang[-2:])):
        model = nlp
    if model is None and lang:
        try:
            model = spacy.load(lang)
        except OSError:
            spacy.cli.download(lang)
            model = spacy.load(lang)
    if model is None:
        if not lang:
            for lang in LANGS:
                try:
                    model = spacy.load(lang)
                    break
                except OSError:
                    pass
        else:
            if model is None:
                lang = LANG
                spacy.cli.download(lang)
                model = spacy.load(lang)
    log.info(
        f"Finished loading SpaCy model: {model} ({model.meta['lang']}_{model.meta['name']})\n"
        f"    with {model.meta['accuracy']['token_acc']:.2f}% token accuracy\n"
        f"     and {model.meta['accuracy']['ents_f']:.2f}% named entity recognition F1 score.\n"
    )
    model = add_hunspell_pipe(model)
    if nlp is None:
        nlp = model
    # load the highest accuracy model into the global singleton nlp variable (user can still override)
    if nlp.lang == 'en':
        if nlp.meta['accuracy']['token_acc'] < model.meta['accuracy']['token_acc']:
            nlp = model
    UNKNOWN_WORDVEC = np.random.randn(nlp._meta['vectors']['width'])
    UNKNOWN_WORDVEC /= np.linalg.norm(UNKNOWN_WORDVEC)
    return model  # return value may be lower accuracy, so `nlp=load('en_web_core_sm')` will have lower accuracy `nlp`


# FIXME: doesn't inherit from spacy.nlp so needs to be deleted in favor of load() function above
class SpacyLM:
    nlp = None
    lang = None

    def __init__(self, lang=None, eager=False):
        self.lang = lang
        if eager:
            self.nlp = load(self.lang)

    def __call__(self, *args, **kwargs):
        self.lang = kwargs.pop('lang', self.lang)
        if self.nlp is None:
            self.nlp = load(self.lang)
            # self.nlp.on_load()
        return self.nlp(*args)


nlp = load(LANG)



import time

from qary.etl.utils import md5


def test_hash_speed():
    runtime = 0
    for i in range(10):
        t0 = time.time()
        for j in range(1000):
            md5('barackobama' + str(j), 16)
        t1 = time.time()
        runtime += (t1 - t0) / 1000 / 10
    assert runtime < 1e-4
    return runtime


def test_hash_consistency():
    assert md5("hello world") != md5("hello worlD")
    assert md5("hello world") != md5("Hello world")
    assert md5("hello world") == md5(b'hello world')
    assert md5("Barack Obama") == 4560975168341313090
    assert md5("barack obama") == 4205506257154982495


if __name__ == '__main__':
    test_hash_consistency()
    test_hash_speed()



""" Transformer based chatbot dialog engine for answering questions """

import logging
import os
import urllib.request
import uuid
import zipfile
from pathlib import Path
import numpy as np
from multiprocessing import cpu_count

from qary.qa.torch_models import QuestionAnsweringModel
from qary.constants import DATA_DIR, USE_CUDA, MIDATA_URL, MIDATA_QA_MODEL_DIR, args
from qary.skills.base import ContextBot
from qary.etl.netutils import DownloadProgressBar
from qary.etl import scrape_wikipedia

log = logging.getLogger(__name__)

STOP_WIKI_PROBABILITY = .4


class Skill(ContextBot):
    """ Skill that provides answers to questions given context data containing the answer """

    def __init__(self, context=None, args=args, **kwargs):
        context = {'doc': {'text': ''}}
        super().__init__(context=context, args=args, **kwargs)
        log.debug(f'Initial qa.Skill.context: {self.context}')
        self.transformer_loggers = []
        for name in logging.root.manager.loggerDict:
            if (len(name) >= 12 and name[:12] == 'transformers') or name == 'qary.skills.qa.utils':
                self.transformer_loggers.append(logging.getLogger(name))
                self.transformer_loggers[-1].setLevel(logging.ERROR)

        qa_model = args.qa_model
        url_str = f"{MIDATA_URL}/{MIDATA_QA_MODEL_DIR}/{qa_model}.zip"
        log.debug(f"Attempting to download url: {url_str}")
        model_dir = os.path.join(DATA_DIR, 'qa-models', f"{qa_model}")
        model_type = qa_model.split('-')[0].lower()
        self.wikiscraper = scrape_wikipedia.WikiScraper()
        if not os.path.isdir(model_dir):
            os.makedirs(model_dir)

        required_files = 'config.json pytorch_model.bin tokenizer_config.json version.json vocab.txt'.split()
        if model_type == 'albert':
            required_files[-1] = 'spiece.model'
        if not all(Path(model_dir, fn).is_file() for fn in required_files):
            zip_local_path = os.path.join(DATA_DIR, 'qa-models', f"{qa_model}.zip")
            with DownloadProgressBar(
                    unit='B', unit_scale=True, miniters=1,
                    desc=url_str.split('/')[-1]
            ) as t:
                urllib.request.urlretrieve(url_str, filename=zip_local_path, reporthook=t.update_to)
            with zipfile.ZipFile(zip_local_path, 'r') as zip_file:
                zip_file.extractall(os.path.join(DATA_DIR, 'qa-models'))
            os.remove(zip_local_path)

        model_args = {
            'process_count': cpu_count() - 2 if cpu_count() > 2 else 1,
            'output_dir': model_dir,
            'cache_dir': model_dir,
            'no_cache': True,
            'use_cached_eval_features': False,
            'overwrite_output_dir': False,
            'silent': True
        }

        self.model = QuestionAnsweringModel(
            model_type, model_dir, args=model_args,
            pretrained=True, use_cuda=USE_CUDA
        )

    def encode_input(self, statement, text):
        """ Converts statement and context strings into json format compatible with BERT transformer

        BERT models use the word "context" to mean a single str read by BERT for QA.
        `qary.skills.base` uses context to manage a dict of many variables
        that help manage conversation

        >>> bot = Skill()
        >>> encoded = bot.encode_input('statement', text='example text')
        >>> encoded[0]['qas'][0]['question']
        'statement'
        >>> encoded[0]['context']
        'example text'
        """
        encoded = [{
            'qas': [{
                'id': str(uuid.uuid1()),
                'question': statement
            }],
            'context': text
        }]
        return encoded

    def decode_output(self, output):
        """
        Extracts reply string from the model's prediction output

        >>> bot = Skill()
        >>> bot.decode_output(
        ...    [{'id': 'unique_id', 'answer': 'response', 'probability': 0.75}])
        (0.75, 'response')
        """
        return output[0]['probability'], output[0]['answer']

    def reply(self, statement, context=None, **kwargs):
        """ Use context document + BERT to answer question in statement

        context is a nested dictionary with two ways to specify the documents for a BERT context:
        {docs: ['doc text 1', '...', ...]}
        or, as only syntactic sugar (internally this is converted to the format above)
        {doc: {text: 'document text ...'}}
        """
        log.info(f"ContextBot.reply(statement={statement}, context={context})")
        if isinstance(context, str):
            context = dict(docs=[context], doc=dict(text=context))
        elif isinstance(context, (list, tuple, np.ndarray)):
            context = dict(docs=context, doc=dict(text='\n'.join(context)))
        # this calls self.update_context(context=context) internally:
        elif isinstance(context, dict):
            if not context.get('docs') and context.get('doc'):
                context['docs'] = [context.get('doc').get('text')]

        if (not context or not any(context.get('docs', ['']))):
            gendocs = self.wikiscraper.find_article_texts(
                query=statement,
                max_articles=1, max_depth=1,
                ngrams=3,
                ignore='who what when where why'.split()) or []
            context = dict(docs=gendocs)
        responses = super().reply(
            statement=statement, context=context, **kwargs) or []
        log.info(f"qa_bots.Skill.super() responses (before BERT): {responses}")
        docs = self.context.get('docs') or [self.context['doc']['text']]
        for i, text in enumerate(docs):
            log.info(
                f"text[{i}] from context['doc']['text'] or scrape_wikipedia: "
                f"{repr(text)[:40]}...{repr(text)[-40:]} ({len(text)} chars)")
            super().update_context(context=dict(doc=dict(text=text)))
            if len(text.strip()) < 2:
                log.info(f'Context document text was too short: "{text}"')
                continue
            encoded_input = self.encode_input(statement, text)
            encoded_output = self.model.predict(encoded_input)
            probability, response = self.decode_output(encoded_output)
            if len(response) > 0:
                responses.append((probability, response))
                if probability > STOP_WIKI_PROBABILITY:
                    log.info(f"Short circuiting wiki crawl because p > thresh: {probability} > {STOP_WIKI_PROBABILITY}")
                    break
        return responses



""" Pattern and template based chatbot dialog engines """
import re


# from .template_generators import generate_sentence  # noqa


class Skill:
    def reply(self, statement, context=None):
        r""" Chatbot "main" function to respond to a user command or statement

        >>> bot = Skill()
        >>> bot.reply('Hi')[0][1]
        'Hello!'
        >>> len(bot.reply('Hey Mycroft!'))
        5
        """
        responses = []
        match = re.match(r'\b(hi|hello|hey)\b(.*)', statement.lower())
        if match:
            responses.append((0.1, "Hello"))
            if 'mycroft' in match.groups()[1].lower():
                responses.append((0.2, "Hi!"))
                responses.append((0.2, "Hi! What would you like to talk about?"))
                responses.append((0.2, "Hi! You remembered me! How are you doing?"))
            if 'bot' in match.groups()[1].lower():
                responses.append((0.2, "Hey. That's a good one."))
        responses.append((0.05, "Wuh?"))
        if statement == 'Hi':
            responses = [(1.0, "Hello!")]
        return responses



""" Transformer based chatbot dialog engine for answering questions """
import logging
import os
import uuid
# import urllib.request
import zipfile
# from multiprocessing import cpu_count

from qary.etl.netutils import download_if_necessary
from qary.constants import MIDATA_URL, MIDATA_QA_MODEL_DIR, args  # , USE_CUDA
from qary.etl.fileutils import LARGE_FILES
from qary.etl.nesting import dict_merge


log = logging.getLogger(__name__)


def normalize_replies(replies=''):
    """ Make sure a list of replies includes score and text

    >>> normalize_replies(['hello world'])
    [(1e-10, 'hello world')]
    """
    if isinstance(replies, str):
        replies = [(1e-10, replies)]
    elif isinstance(replies, tuple) and len(replies, 2) and isinstance(replies[0], float):
        replies = [(replies[0], str(replies[1]))]
    # TODO: this sorting is likely unnecessary, redundant with sort happening within CLIBot.reply()
    return sorted([
        ((1e-10, r) if isinstance(r, str) else tuple(r))
        for r in replies
    ], reverse=True)


class Replies(list):
    """ WIP: Container for 2-tuples of scored replies (strings)

    TODO:
        - override __getitem__ to return None when empty
        - validate method to make sure 2-tuples of float, str
        - normalize method to ensures scores add up to 1
        - softmax to improve likelihood of best replies
        - sort method?
    """
    pass


class ContextDict(dict):
    """ WIP: Container for context dict attributes for bots

    ContextDict.merge should use dict_merge
    """

    def merge(self, other):
        dict_merge(self, other)

    def dict_merge(self, other):
        dict_merge(self, other)


class EmptyRepliesBot:
    # def __init__(self, **kwargs):
    #     super().__init__()

    def reply(self, statement, context=None):
        """ Chatbot "main" function to respond to a user command or statement

        >>> bot = EmptyRepliesBot()
        >>> bot.reply('Hi')
        []
        """
        return []


class HiBot(EmptyRepliesBot):
    def reply(self, statement, context=None):
        """ Chatbot "main" function to respond to a user command or statement

        >>> bot = HiBot()
        >>> bot.reply('Hi')
        [(1.0, 'Hi!')]
        """
        s = statement.lower().strip().split()
        if s and s[0] in ('hi', 'hello', 'howdy', 'helo', 'yo'):
            return [(1.0, 'Hi!')]
        return []


class ContextBot:
    """ Manages self.context attribute with update_, reset_ and reply(context=...) methods

    >>> conbot = ContextBot({'init': 'init_value'})
    >>> conbot.context
    {'args': Namespace(bots=...words=[]), 'init': 'init_value'}
    >>> conbot.update_context({'new': 'new_value', 'args': None})
    {'args': None, 'init': 'init_value', 'new': 'new_value'}
    >>> conbot.context
    {'args': None, 'init': 'init_value', 'new': 'new_value'}
    >>> conbot.update_context({'init': 'updated_existing', 'new': {'inner': 'new_innards'}})
    {'args': None, 'init': 'updated_existing', 'new': {'inner': 'new_innards'}}
    >>> conbot.context
    {'args': None, 'init': 'updated_existing', 'new': {'inner': 'new_innards'}}
    >>> conbot.reply('Hi', context={'new': {'inner_reply': 'new_inner_reply'}})
    []
    >>> conbot.context['init']
    'updated_existing'
    >>> conbot.context['new']
    {'inner': 'new_innards', 'inner_reply': 'new_inner_reply'}
    """

    def __init__(self, context=None, args=args, **kwargs):
        super().__init__(**kwargs)
        self.args = args
        self.context = {'args': self.args}
        # self.context = {}
        if isinstance(context, str):
            log.warning("Deprecated API: `context` should be a nested dictionary. Texts belong at `context['doc']['text']`).")
            self.update_context({'doc': {'text': context}})
        elif isinstance(context, dict):
            self.update_context(context)

    def update_context(self, context=None):
        log.debug(f"Updating self.context using context: {repr(context)[:60]}")
        if isinstance(context, str):
            context = {'doc': {'text': context}}
        context = {} if context is None else dict(context)
        dict_merge(self.context, context)
        log.debug(f"Updated self.context: {repr(self.context)[:60]}")
        return self.context

    def reset_context(self, context=None):
        self.context = {'args': self.args}
        log.info(f"Reset self.context to self.context={repr(self.context)[:60]} before updating with {repr(context)[:60]}")
        self.update_context(context=context)
        return self.context

    def reply(self, statement, context=None):
        """ Chatbot "main" function to respond to a user command or statement

        >>> bot = ContextBot()
        >>> bot.reply('Hi', context={'new': 'context'})
        []
        >>> bot.context['new']
        'context'
        >>> len(bot.context.keys())
        2
        """
        self.update_context(context=context)
        return []


class HistoryBot:
    """ Remembers the history of every user statement and bot response string

    >>> histbot = HistoryBot()
    >>> histbot.history
    []
    >>> histbot.update_history('Hi')
    1
    >>> histbot.history
    [{'statement': 'Hi', 'possible_statements': (), 'agent': None}]
    >>> histbot.update_history('Hello', agent='human')
    2
    >>> histbot.history
    [{'statement': 'Hi', 'possible_statements': (), 'agent': None},
     {'statement': 'Hello', 'possible_statements': (), 'agent': 'human'}]
    """

    def __init__(self, history=None, **kwargs):
        super().__init__(**kwargs)
        self.history = history or []

    def update_history(self, statement=None, possible_statements=None, agent=None):
        """ Append convo history with a statement/reply, the possibilities it was chosen from, and an agent (bot/user) name """
        self.history.append(dict(
            statement=statement,
            possible_statements=tuple(possible_statements or []),
            agent=agent))
        return len(self.history)

    def reset_history(self, statement=None, possible_statements=None, agent=None):
        """ Reset conversation history to a single statement/reply

        Resets the conversation history to the last interaction:
            - a single most recent reply
            - the possibile Skill replies list that the reply was chosen from
            - the Skills that created the list of possible replies
        """
        self.history = []
        if statement or possible_statements or agent:
            self.update_history(
                statement=statement,
                possible_statements=possible_statements,
                agent=agent)
        return len(self.history)


class TransformerBot(HistoryBot, ContextBot):
    """ Base Skill class that maintains context and load transformer models.  """

    def __init__(self, context=None, args=args, history=None, **kwargs):
        super().__init__(context=context, args=args, history=history, **kwargs)

    def load_model(self, args):
        self.transformer_loggers = []
        for name in log.root.manager.loggerDict:
            if (len(name) >= 12 and name[:12] == 'transformers') or name == 'qary.skills.qa.utils':
                self.transformer_loggers.append(log.getLogger(name))
                self.transformer_loggers[-1].setLevel(log.ERROR)

        qa_model = args.qa_model
        url_str = f"{MIDATA_URL}/{MIDATA_QA_MODEL_DIR}/{qa_model}.zip"
        log.warning(f"Attempting to download url: {url_str}")
        model_dir = os.path.dirname(LARGE_FILES['albert-large-v2']['path'])
        model_type = qa_model.split('-')[0].lower()
        if not os.path.isdir(model_dir):
            os.makedirs(model_dir)

        if not all((
            os.path.exists(os.path.join(model_dir, 'config.json')),
            os.path.exists(os.path.join(model_dir, 'pytorch_model.bin')),
            os.path.exists(os.path.join(model_dir, 'tokenizer_config.json')),
            os.path.exists(os.path.join(model_dir, 'version.json')),
            any((model_type == 'bert' and os.path.exists(os.path.join(model_dir, 'vocab.txt'))),
                (model_type == 'albert' and os.path.exists(os.path.join(model_dir, 'spiece.model')))),
        )):
            zip_local_path = download_if_necessary(
                url=LARGE_FILES['albert-large-v2']['url'],
                path=LARGE_FILES['albert-large-v2']['path'])
            with zipfile.ZipFile(zip_local_path, 'r') as zip_file:
                zip_file.extractall(os.path.dirname(LARGE_FILES['albert-large-v2']['path']))
            os.remove(zip_local_path)

    def encode_transformer_input(self, statement, context=None):
        """ Convert statement and context strings into nested dict format compatible with [AL]BERT transformer

        >>> bot = TransformerBot()
        >>> encoded = bot.encode_transformer_input('statement', 'context')
        >>> encoded[0]['qas'][0]['question']
        'statement'
        >>> encoded[0]['context']
        'context'
        """
        if context is None:
            context = self.context
        if isinstance(context, str):
            # TODO try/except on [] instead of get to provide deprecation warnings
            text = context
            super().update_context({'doc': {'text': text}})
        else:
            super().update_context(context)
        text = self.context['doc']['text']
        encoded = [{
            'qas': [{
                'id': str(uuid.uuid1()),
                'question': statement
            }],
            'context': text}]
        return encoded

    def decode_transformer_output(self, output):
        """ Extract reply string from the transformer model's prediction output (nested dict)

        >>> bot = TransformerBot()
        >>> bot.decode_transformer_output([{'id': 'unique_id', 'answer': 'response', 'probability': 0.75}])
        (0.75, 'response')
        """
        return output[0]['probability'], output[0]['answer']



""" Text search to retrieve a statement """
import logging
import sys
import os

import yaml
import pandas as pd
from rapidfuzz import process
from qary.constants import DATA_DIR

log = logging.getLogger(__name__)


LIMIT = 1000000


def get_data(name):
    """ replacement for nlpia.loaders.get_data to avoid dependencies that result in version conflicts

    >>> get_data('movie_dialog').shape
    (64350, 2)
    """
    return pd.read_csv(os.path.join(DATA_DIR, name + '.csv'))


def normalize(text):
    return text.lower()


def scale_probability(p):
    """ Levenshtein similarity is only good when it's high, when it's low, the score should go down """
    return p ** 2


def load_faq(faq_path=os.path.join(DATA_DIR, 'dsfaq_plus_faq_data_science_and_machine_learning.yml')):
    faq = None
    with open(faq_path, 'r') as instream:
        try:
            faq = yaml.safe_load(instream)
        except yaml.YAMLError as e:
            print(e)
            raise(e)
    for i, qa in enumerate(faq):
        if not isinstance(qa, dict):
            faq[i] = {}
            log.warning(f'qa #{i} was not a dict')
            continue
        for k in qa:
            if k.lower() != k:
                qa[k.lower()] = qa.pop(k)
        # if 'q' not in qa:
        #     log.warning(f'qa #{i} had no Question: {list(qa)} {qa[list(qa)[0]]}')
        #     qa['q'] = qa.pop('q_student', qa.pop('q_student2', qa.pop('q_teacher')))
        # if 'a' not in qa:
        #     log.warning(f'qa #{i} had no Answer: {list(qa)} {qa[list(qa)[0]]}')
        #     qa['a'] = qa.pop('a_teacher', qa.pop('a_teacher2', qa.pop('a_student')))
        #     continue
    faq = pd.DataFrame(faq)
    faq = faq.dropna()
    return faq


class Skill:
    db = None

    def __init__(self, name='movie_dialog'):
        self.limit = LIMIT
        self.name = name
        # TODO: make this lazy, do it inside reply()
        self.db = self.load_dialog(name=name)

    def load_dialog(self, name='movie_dialog'):
        """ Load movie dialog database

        >>> bot = Skill()
        >>> db = bot.load_dialog()
        >>> len(db)
        58007
        """
        log.warning('Loading movie dialog...')
        if name == 'dsfaq':
            db = load_faq()
        else:
            db = get_data(name)
        log.info(f'Loaded {len(db)} {self.name} statement-reply pairs.')
        if self.limit <= len(db):
            log.info(f'Limiting {self.name} database to {self.limit} statement-reply pairs.')
            db = db.iloc[:self.limit]
        db = dict(zip(db[db.columns[0]], db[db.columns[1]]))
        return db

    def reply(self, statement, context=None, db=None):
        """ Use rapidfuzz to find the closest key in the dictionary then return the value for that key

        >>> bot = Skill()
        >>> reply = bot.reply

        >> reply('hey', {'hello': 'world', 'goodbye': 'fantasy'})
        (0.3, 'fantasy')
        >> reply("Hi!")
        (1.0, 'hey there. tired of breathing?')
        """
        if self.db is None:
            self.db = db
        if self.db is None:
            self.db = self.load_dialog()
        movie_statement, percent_match, movie_reply = process.extractOne(
            normalize(statement), choices=self.db)
        log.info(f'Closest movie_statement = {movie_statement}')
        return [((percent_match / 100.), movie_reply)]


# class FAQBot(Skill):
#     def __init__(self, name='dsfaq'):
#         self.limit = LIMIT
#         # TODO: make this lazy, do it inside reply()
#         self.db = self.load_dialog(name=name)
# BOTS = (Skill, FAQBot)

if __name__ == '__main__':
    if len(sys.argv) > 1:
        bot = Skill()
        statement = "Hi!"
        statement = ' '.join(sys.argv[1:])
        print(bot.reply(statement))



from simpletransformers.question_answering import QuestionAnsweringModel
import json
import os
import logging

from tqdm import tqdm

from qary.constants import DATA_DIR
log = logging.getLogger(__name__)


SQUAD_TRAINING_JSON = os.path.join(DATA_DIR, 'simple-transformers', 'train-v2.0.json')
ST_TRAINING_JSON = os.path.join(DATA_DIR, 'simple-transformers', 'squad-train.json')


def preprocess_squad(squad_path=SQUAD_TRAINING_JSON):
    """ Load raw SQUAD training set v2.0 from Stanford format json and convert to simpletransformer format

    Cache the simpletransformer json format, if it doesn't yet exist in data/simple-transformer
    or the provided path
    """
    with open(squad_path, 'r') as fin:
        squad_train = json.load(fin)
    log.info(f"Loaded {len(squad_train['data'])} examples from {squad_path} dataset...")
    return [item for topic in tqdm(squad_train['data']) for item in topic['paragraphs']]


def load_trainset(train_path=ST_TRAINING_JSON, squad_path=SQUAD_TRAINING_JSON):
    """ Load a preprocessed SQUAD training set into RAM as a list or create it using the raw squad dataset """
    try:
        with open(train_path, 'r') as fin:
            trainset = json.load(fin)
    except (IOError, FileNotFoundError):
        trainset = preprocess_squad(squad_path)
        with open(train_path, 'w') as fout:
            json.dump(trainset, fout)
    return trainset


def train_squad(train_path=ST_TRAINING_JSON, squad_path=SQUAD_TRAINING_JSON):
    trainset = load_trainset(train_path, squad_path=squad_path)
    model = QuestionAnsweringModel('distilbert', 'distilbert-base-uncased-distilled-squad', use_cuda=False,
                                   args={'reprocess_input_data': True, 'overwrite_output_dir': True})
    model.train_model(trainset)

    return model


def evaluate(model, test_path=ST_TRAINING_JSON):
    # Evaluate the model. (Being lazy and evaluating on the train data itself)
    trainset = load_trainset(test_path)

    result, text = model.eval_model(trainset)

    log.info(result)
    log.info(text)
    return result, text


if __name__ == '__main__':
    model = train_squad()
    result, text = evaluate(model)



""" Transformer based chatbot dialog engine for answering questions """

import logging

from qary.skills import qa

log = logging.getLogger(__name__)

STOP_WIKI_PROBABILITY = .4


class Skill(qa.Skill):
    """ Skill that provides answers to questions given context data containing the answer """

    def __init__(self, context=None, **kwargs):
        super().__init__(context=context, **kwargs)



""" Functions, regular expressions, and SpaCy patterns for extracting information and knowledge from natural language text """
import re
from qary.spacy_language_model import nlp


def whatis(statement):
    """ Extract the target noun phrase phrase for "What is ... " questions.

    >>> whatis("What is a Secretary Bird?")
    'Secretary Bird'
    """
    match = re.match(
        pattern=r"\b(what\s+(is|are)\s*(not|n't)?\s+(a|an|the)?)\s*\b([^\?]*)(\?*)",
        string=statement.strip(),
        flags=re.IGNORECASE)
    extracted_term = match.groups()[-2] if match else ''
    return extracted_term.strip()


def whatmeans(statement):
    """ Extract the target noun phrase for questions like "What does RMSE mean?"

    >>> whatmeans("What does rmse mean?")
    'rmse'
    """
    pattern = (r"\b(what\s+(is|are)\s*(not|n't)?\s+(a|an|the|those|this|that)?)\s*\b"
               r"((meaning|definition|significance|origin|interpretation|translation)[s]*\s*|"
               r"((mean|define|signifie|originate|interpret|translate)[s]*\s*))"
               r"(of|for|to|in|on)?\s*((a|an|the|those|this|that)?[s]*)?\s*"
               r"((tech|technology|science|biology|medicine|health|healthcare|genomic|data[ ]?science|machine[ ]?learning|dl|ml|ds|it|is|ir|nlp)?[es])*\s*"
               r"((phrase|quote|term|terminology|word|token|n[-]?gram?)[es]*)?\s*"
               r"([^\?]*)(\?*)"
               )
    match = re.match(
        pattern=pattern,
        string=statement.strip(),
        flags=re.IGNORECASE)
    extracted_term = match.groups()[-2] if match else ''
    if not extracted_term:
        match = match or re.match(
            pattern=(r"\b(what\s+(does)?\s*(a|an|the)?)\s*"
                     r"([^\?\s]+\s+)+"
                     r"\b(mean|signify|represent|translate|stand)?\s*(for|to|as|in)?\s*(\?*)"),
            string=statement.strip(),
            flags=re.IGNORECASE)
        extracted_term = match.groups()[-4] if match else ''
    return extracted_term.strip()


def count_upper(s):
    return sum(1 for c in s if c.isupper())


def count_lower(s):
    return sum(1 for c in s if c.isupper())


def isacronym(term):
    """ Estimate the probability that a given term is an acronym based only on its capitalization

    >>> isacronym('ABC')
    0.6666...
    >>> isacronym('Cat')
    0.0
    >>> isacronym('Cy')
    0.0
    """
    if re.match(r'\s', term):
        return 0
    return min(max(
        sum(1 if c.isupper() else -1 if c.islower() else -2 for c in term) / len(term) - .6, 0) / .6, 1)


def extract_name(s):
    """ from a sentence identify and extract proper nouns

    >>> extract_name("Who is Elon Musk?")
    'Elon Musk'
    """
    name = []
    for word in nlp(s):
        if word.pos_ == "PROPN":
            name.append(word.text)
    return " ".join(name)



# template_generators
import re

import numpy as np

from .constants import SENTENCE_SPEC
from .spacy_language_model import nlp


def generate_sentence(spec=SENTENCE_SPEC, sentence_id=None):
    """ Generate random sentence using word probabilities specified in SENTENCE_SPEC

    >>> spec = {
    ...     "answers":[[{"HDL":0.95,"good_cholesterol":0.05}, {"150": 0.01,"145": 0.01,"unk": 0.98}],
    ...     "sentences":["Patient LDL level is 100, ________ level is 50, and the total is ______ .",]
    ...     }
    >>> s = generate_sentence(spec=spec, sentence_id=0)  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
    >>> s
    'Patient LDL level is 100, ... level is 50, and the total is ... .'
    >>> s[26:42] in ('HDL level is 50,', 'good_cholesterol')
    True
    >>> s[60:63] in ('150', '145', 'unk')
    True
    """
    sentences = spec['sentences']
    if sentence_id is None:
        sentence_id = np.random.randint(0, len(sentences))
    sentence = sentences[sentence_id]
    answer = spec['answers'][sentence_id]
    i_unk = 0
    tokens = []
    for i, tok in enumerate(nlp(sentence)):
        if re.match(r'^(_+|unk|\[MASK\])$', tok.text):
            possible_tokens, p = list(zip(*answer[i_unk].items()))
            tokens.append(np.random.choice(a=possible_tokens, p=p))
            i_unk += 1



import logging
import nltk
from qary import spacy_language_model

log = logging.getLogger(__name__)
nlp = spacy_language_model.nlp

try:
    assert nlp._meta['vectors']['width'] == 300  # len(nlp('word vector').vector) < 300:
except AssertionError:
    log.warning(f"SpaCy Language model ({nlp._meta['name']}) doesn't contain 300D word2vec word vectors.")
    nlp = spacy_language_model.nlp = spacy_language_model.load('en_core_web_md')
assert nlp._meta['vectors']['width'] == 300


def iou(a, b):
    """ Crude character vector overlap measure of string similarity

    >>> iou('Hello', 'World')
    0.285...
    """
    a, b = set(a.lower().strip()), set(b.lower().strip())
    return len(a & b) / len(a | b)


def bleu(text, reference_text, weight1=0.25, weight2=0.25, weight3=0.25, weightL=0.25):
    """Compute the BLEU similarity between a text string and a reference text.

    All the different weights are for different N-grams. E.g.weight1 for unigram.
    Both the texts must have at least four words. Otherwise the BLEU score is zero.
    The full-stops matter
    >>> bleu('I visited the park and enjoyed.', 'I visited the park and had fun ')
    0.6148
    >>> bleu('','')
    0
    >>> bleu('','I am happy')
    0
    >>> bleu('I visited', 'I visited the park and had fun ')
    0.0
    >>> bleu('I am happy and contented', 'I am happy and joyous')
    0.6687
    >>> bleu('I am going to visit the park and have fun.','I am going to visit.')
    0.3672
    >>> bleu('I am going to visit the park and have fun.','I am going to visit')
    0.3508
    """

    doc = nlp(text)
    reference_doc = nlp(reference_text)

    tokens = [token.text for token in doc]
    reference_tokens = [token.text for token in reference_doc]

    bleu_score = nltk.translate.bleu_score.sentence_bleu([reference_tokens], tokens, weights=(weight1, weight2, weight3, weightL))
    return round(bleu_score, 4)


def overlapping_words(tokens, reference_tokens):
    count = 0
    extra_list = []
    for word1 in tokens:
        if word1 in reference_tokens:
            count += 1
            extra_list.append(word1)
    for word2 in reference_tokens:
        if word2 not in extra_list and word2 in tokens:
            count += 1
    return count


def rouge1(text, reference_text):
    """Compute the ROUGE similarity between a text string and a reference text with unigram.

    Case doesn't matter, full-stop doesn't matter.
    Both the texts can be anything except blanks and punctuations as long as the reference text has it.
    A character, number can also be considered as a word if it's in the reference text.
    >>> rouge1('police killed the gunman', 'police kill the gunman')
    1.0
    >>> rouge1('police killed the gunman', 'the gunman was killed by the police')
    0.7273
    >>> rouge1('police killed the gunman', 'gunman killed the police')
    1.0
    >>> rouge1('police killed the gunman', 'police is killing the gunman')
    0.8889
    >>> rouge1('police killed the gunman.', 'police killed the gunman')
    1.0
    >>> rouge1('police killed the gunman', 'POLICE killed the gunman')
    1.0
    >>> rouge1('police killed the gunman', 'police killed the')
    0.8571
    >>> rouge1('police killed the gunman', 'police killed')
    0.6667
    >>> rouge1('police killed the gunman', '')
    0.0
    >>> rouge1('police killed the gunman', 'police')
    0.4
    >>> rouge1('', '')
    0.0
    >>> rouge1('?', '?')
    0.0
    """
    stemmer = nltk.stem.PorterStemmer()
    doc = nlp(text)
    reference_doc = nlp(reference_text)
    tokens = [token.text for token in doc]
    reference_tokens = [token.text for token in reference_doc]
    punctuations = '''!()-[]{};:'"r"\",<>./?@#$%^&*_~'''
    tokens = [stemmer.stem(word.lower()) for word in tokens if word not in punctuations]
    reference_tokens = [stemmer.stem(word.lower()) for word in reference_tokens if word not in punctuations]
    count = overlapping_words(tokens, reference_tokens)
    try:
        recall = count / len(reference_tokens)
    except ZeroDivisionError:
        recall = 0.0
    try:
        precision = count / len(tokens)
    except ZeroDivisionError:
        precision = 0.0
    try:
        fmeasure = 2 * ((precision * recall) / (precision + recall))
    except ZeroDivisionError:
        fmeasure = 0.0
    return round(fmeasure, 4)


def score(reply, stmt=None, **kwargs):
    """ Compute word2vec docvec cosine similarity (fall back to character IOU)

    >>> score('Hello world!', 'Goodbye big earth!') > .5
    True
    """
    global nlp
    nlp = kwargs.get('nlp', nlp)
    if kwargs is None or nlp is None or not stmt or not reply:
        return 0.0

    reply_doc, stmt_doc = nlp(str(reply)), nlp(str(stmt))

    if not reply_doc or not stmt_doc or not reply_doc.has_vector or not stmt_doc.has_vector:
        # FIXME: levenshtien would be better or fuzzywuzzy
        return iou(reply, stmt)

    cos_sim = reply_doc.similarity(stmt_doc)
    log.debug(f'cos_sim={cos_sim}')
    return cos_sim


class Doc:
    global nlp

    def __init__(self, text='', nlp=nlp):
        """ Create a Doc object with an API similar to spacy.Doc

        >>> d = Doc('Hello')
        >>> len(d.vector)
        300
        >>> d.doc.similarity(d.doc) > .99
        True
        """
        self.nlp = nlp if nlp is not None else self.nlp
        self.text = text
        self.doc = nlp(text)
        self.vector = self.doc.vector

    def similarity(self, other_doc):
        """ Similarity of self Doc object meaning to the meaning of another Doc object

        >>> doc = Doc('USA')
        >>> doc.similarity(Doc('United States'))
        0.5...
        """
        if hasattr(other_doc, 'vector_norm'):
            return self.doc.similarity(other_doc)
        else:
            return self.doc.similarity(getattr(other_doc, 'doc', other_doc))


def similarity(text1, text2):
    """ Similarity between two natural language texts (words, phrases, documents) 1 = 100%, -1 = -100%

    >>> similarity('Hello', 'hello') > 0.99
    True
    >>> .8 > similarity('Hello!', 'Hi?') > 0.75
    True
    """
    return Doc(text1).similarity(Doc(text2).doc)

    # log.debug(f"vector1 for text1 {vector1}")
    # question_vector /= np.linalg.norm(question_vector)
    # log.debug(f"faq['question_vectors'].shape is {self.faq['question_vectors'].shape}")
    # question_similarities = self.faq['question_vectors'].dot(question_vector.reshape(-1, 1))



""" Reads files and indexes them for semantic search

- spacy doc vectors for each file
- spacy doc vectors for each sentence in each file

"""

import os
import logging

from tqdm import tqdm
import numpy as np
import pandas as pd
from annoy import AnnoyIndex

from qary.spacy_language_model import nlp
from qary.etl.fileutils import find_files
# from qary.futil import find_files

log = logging.getLogger(__name__)


def get_files(path, size_limit=100000, excludes=['.DS_Store']):
    allfiles = find_files(path)
    df = pd.DataFrame(allfiles)
    df['accessed'] = pd.to_datetime(df.accessed)
    df['modified'] = pd.to_datetime(df.modified)
    df['changed_any'] = pd.to_datetime(df['changed_any'])
    df['ext'] = df.name.apply(lambda s: os.path.splitext(s)[1])
    df['basename'] = df.name.apply(lambda s: os.path.splitext(s)[0])
    df.index = list(range(len(df)))
    df.index.name = 'file_id'
    return df


def is_journal(df, extensions='txt rst md text'.split() + [''], size_limit=100000, excludes=['.DS_Store']):
    extensions = [extensions] if isinstance(extensions, str) else extensions
    extensions = [ext.lower().lstrip('.') for ext in extensions]
    mask = pd.Series([False] * len(df), index=df.index)
    for ext in extensions:
        mask = mask | (df['ext'].str.lower().str.lstrip('.') == ext)
    for excl in excludes:
        mask = mask & (df['name'].str.strip() != excl)
    if size_limit:
        mask = mask & (df['size'] <= size_limit)

    return mask


def get_sentences(df, size_limit=50000, vector_dim=None):
    vector_dim = len(nlp('word').vector) if vector_dim is None else int(vector_dim)
    sents = []
    docvecs = np.zeros((len(df), vector_dim))
    encodings = []
    for file_id, row in tqdm(df.iterrows(), total=len(df)):
        sentvecs = []
        encodings.append('utf8')
        if row['size'] <= size_limit and row['is_journal']:
            try:
                with open(row['path'], 'rb') as fin:
                    bintext = fin.read()
                try:
                    text = bintext.decode()
                except UnicodeDecodeError:
                    encodings[-1] = 'latin'
                    log.warning(f"LATIN?: {row['path']}")
                    text = bintext.decode('latin')
                doc = nlp(text)
            except UnicodeDecodeError:
                log.error(f"UnicodeDecodeError: {row['path']}")
                continue
            docvecs[file_id, :] = np.array(list(doc.vector))
            docsents = [dict(sentence_pos=f'{file_id}-{j}', file_id=file_id, text=s.text) for j, s in enumerate(doc.sents)]
            log.info(f"Read {len(docsents)} sentences: {row['path']}")
            # print(doc.vector)
            sents.extend(docsents)
            sentvecs.extend([s.vector for s in doc.sents])
        else:
            log.warning(f"skipped {row['path']}")

    df['encoding'] = encodings
    df = pd.concat([df, pd.DataFrame(np.array(docvecs))], axis=1)
    df_sents = pd.DataFrame(sents, index=list(range(len(sents))))
    df_sents = pd.concat([df_sents, pd.DataFrame(np.array(sentvecs))], axis=1)
    df_sents.index.name = 'sentence_id'
    return df, df_sents


def index_csv(csv_path='/midata/private/journal/files.csv', num_dims=300):
    df = pd.read_csv(csv_path, index_col=0)
    vectors = df.iloc[:, -num_dims:].values
    num_vecs, num_dims = vectors.shape
    index = AnnoyIndex(f=num_dims)
    for i, vec in tqdm(enumerate(vectors)):
        # if not i % 10000:
        #     print('{}: {}: {}'.format(i, df_notes['name'].iloc[i], vec))
        index.add_item(i, vec)
    num_trees = int(np.log(num_vecs).round(0)) + 1
    index.build(num_trees)
    index.save(os.path.join(os.path.dirname(csv_path), 'files_index.ann'))
    return index


def search_csv(csv_path='/midata/private/journal/files.csv',
               query='Misima island port harbor derelict ship PNG Papua New Guinnea Australis '
                     'harbor storm sailing cliffs anchor drag',
               num_results=10, num_dims=300):
    df = pd.read_csv(csv_path, index_col=0)
    index_path = os.path.join(os.path.dirname(csv_path), 'files_index.ann')
    index = AnnoyIndex(f=num_dims)
    index.load(index_path)
    vec = nlp(query).vector
    paths = []
    for i in index.get_nns_by_vector(vec, num_results):
        path = df.iloc[i]['path']
        paths.append(path)
        print(path)
        with open(path, 'rb') as fin:
            bintext = b''.join(fin.readlines()[:10])
        try:
            text = bintext.decode()
        except UnicodeDecodeError:
            text = bintext.decode('latin')
        print(text)
        print('-' * 120)
    return


# import numpy as np
# num_vectors=len(wv.vocab)
# num_trees=int(np.log(num_vectors).round(0))
# index.build(num_trees)  # <1>
# index.save('Word2vec_index.ann')  # <2>


def files_to_csvs():
    df = get_files('~/Dropbox/notes/journal')
    df['is_journal'] = is_journal(df)
    df, df_sents = get_sentences(df)
    df.to_csv('files.csv')
    df_sents.to_csv('sentences.csv')


if __name__ == '__main__':
    pass



"""Open Street Maps -- GIS search, edit, scrape, parse

>>> txt = '''
... Wednesday, 3-9pm
... Thursday, 3-9pm
... Friday 3-10pm
...
... Saturday, 12-10pm
... Sunday, 12-6pmThursday, 3-9pm
... '''
>>> sorted(list(extract_opening_hours(txt).keys()))
['Fr', 'Sa', 'Su', 'Th', 'We']
"""
from datetime import time


def extract_opening_hours(s):
    # FIXME: parse s
    return dict(
        We=(time(15), time(21)),
        Th=(time(15), time(21)),
        Fr=(time(15), time(22)),
        Sa=(time(12), time(22)),
        Su=(time(12), time(21)))


def to_osm_format(data):
    """ OSM "opening_hours" tag field format

    SEE: https://wiki.openstreetmap.org/wiki/Key:opening_hours
    """
    txt = ''
    for dow in 'Su Mo Tu We Th Fr Sa'.split():
        day_hours = data.get(dow, None)
        if day_hours is not None:
            txt += f'{dow} {day_hours[0].hour}:{day_hours[0].minute}-{day_hours[2].hour}:{day_hours[1].minute}; '
    return txt.rstrip().rstrip(';')



""" spell checker bot based on Peter Norvig's spell checker """

from qary.etl.spell import make_spelling_corrections


class Skill:
    def reply(self, statement):
        r""" Chatbot "main" function to respond to a user command or statement

        >>> skill = Skill()
        >>> skill.reply('What is probabillity?')
        [(1.0, '\n Original: What is probabillity?\nCorrected: What is probability?')]
        >>> skill.reply('When was the telefone inventid?')
        [(1.0, '\n Original: When was the telefone inventid?\nCorrected: When was the telephone invented?')]
        >>> skill.reply('When was the telephone invented?')
        [(0.33, "I can't find any misspelled words.")]
        """
        corrected_statement = make_spelling_corrections(statement)
        confidence = 1.0 if corrected_statement.strip().lower() != statement.strip().lower() else 0.33
        if confidence == 1:
            return [(confidence, f"\n Original: {statement}\nCorrected: {corrected_statement}")]
        else:
            return [(confidence, f"I can't find any misspelled words.")]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This is a skeleton file that can serve as a starting point for a Python
console script. To run this script uncomment the following lines in the
[options.entry_points] section in setup.cfg:

    console_scripts =
         bot = qary.skeleton:run

Then run `python setup.py install` which will install the command `fibonacci`
inside your current environment.
Besides console scripts, the header (i.e. until _logger...) of this file can
also be used as template for Python modules.

Note: This skeleton file can be safely removed if not needed!

```bash
$ bot -b 'pattern,parul' -vv -p --num_top_replies=1
YOU: Hi
bot: Hello!
YOU: Looking good!
```
"""
from collections import abc
import importlib
import json
import logging
import os
import sys

import numpy as np
import pandas as pd

from qary import constants
from qary.skills.base import normalize_replies

__author__ = "see AUTHORS.md and README.md: Travis, Nima, Erturgrul, Aliya, Xavier, Maria, Hobson, ..."
__copyright__ = "Hobson Lane"
__license__ = "The Hippocratic License (MIT + *Do No Harm*, see LICENSE.txt)"


log = logging.getLogger(__name__)

log.info(f'sys.path: {sys.path}')
log.info(f'constants: {dir(constants)}')
log.info(f'BASE_DIR: {constants.BASE_DIR}')
log.info(f'SRC_DIR: {constants.SRC_DIR}')
log.info(f'sys.path (after append): {sys.path}')
from qary.scores.quality_score import QualityScore  # noqa

BOT = None


class CLIBot:
    """ Conversation manager intended to interact with the user on the command line, but can be used by other plugins/

    >>> CLIBot(bots='parul,pattern'.split(','), num_top_replies=1).reply('Hi')
    'Hello!'
    """
    bot_names = []
    bot_modules = []
    bots = []

    def __init__(
            self,
            bots=constants.DEFAULT_BOTS,
            num_top_replies=None,
            **quality_kwargs):
        if isinstance(bots, str):
            bots = bots.split(',')
        if not isinstance(bots, abc.Mapping):
            bots = dict(zip(bots, [{}] * len(bots)))
        log.info(f'CLIBot(bots={bots}')
        for bot_name, bot_kwargs in bots.items():
            bot_kwargs = {} if bot_kwargs is None else dict(bot_kwargs)
            self.add_skill(bot_name, **bot_kwargs)
        self.num_top_replies = constants.DEFAULT_CONFIG['num_top_replies'] if num_top_replies is None else min(
            max(int(num_top_replies), 1), 10000)
        self.repliers = [bot.reply if hasattr(bot, 'reply') else bot for bot in self.bots if bot is not None]
        log.debug(f'Loaded skills: {self.bots}')
        self.quality_score = QualityScore(**quality_kwargs)

    def add_skill(self, bot_name, **bot_kwargs):
        log.info(f'Adding bot named {bot_name}')
        self.bot_names.append(bot_name)
        bot_module = importlib.import_module(f'qary.skills.{bot_name}')
        new_modules, new_skills = [], []
        if bot_module.__class__.__name__ == 'module':
            module_vars = tuple(vars(bot_module).items())
            for bot_class in (c for n, c in module_vars if n.endswith('Skill') and callable(getattr(c, 'reply', None))):
                log.info(f'Adding a Skill class {bot_class} from module {bot_module}...')
                new_skills.append(bot_class(**bot_kwargs))
                new_modules.append(bot_module)
        else:
            log.warning(f'FIXME: add feature to allow import of specific bot classes like "qary.skills.{bot_name}"')
        self.bots.extend(new_skills)
        self.bot_modules.extend(new_modules)
        return new_skills

    def log_reply(self, statement, reply):
        history_path = os.path.join(constants.DATA_DIR, 'history.json')
        try:
            history = list()
            with open(history_path, 'r') as f:
                history = json.load(f)
        except IOError as e:
            log.error(str(e))
            with open(history_path, 'w') as f:
                f.write('[]')
        except json.JSONDecodeError as e:
            log.error(str(e))
            log.info('Saving history.json contents to history.json.swp before overwriting')
            with open(history_path, 'r') as f:
                data = f.read()
            with open(history_path + '.swp', 'w') as f:
                f.write(data)
        history.append(['user', statement])
        history.append(['bot', reply])
        with open(history_path, 'w') as f:
            json.dump(history, f)

    def reply(self, statement='', context=None):
        ''' Collect replies from from loaded bots and return best reply (str). '''
        log.info(f'statement={statement}')
        replies = []
        # Collect replies from each bot.
        for bot in self.bots:
            bot_replies = []
            log.info(f'Running bot {bot}')
            # FIXME: create set_context() method on those bots that need it and do away with context try/except
            try:
                bot_replies = bot.reply(statement, context=context)
                log.info("{bot.__name__} replies: {bot_replies}")
            except TypeError as e:
                log.debug(f"TypeError: bot.reply probably got an unexpected keyword argument 'context': {e}")
                bot_replies = bot.reply(statement)
            except Exception as e:
                log.error(f'Error trying to run {bot.__self__.__class__}.reply("{statement}", context={context})')
                log.error(str(e))
                if constants.args.debug:
                    raise e
            log.debug(repr(bot))
            bot_replies = normalize_replies(bot_replies)
            replies.extend(bot_replies)

        # Weighted random selection of reply from those with top n confidence scores
        if len(replies):
            log.info(f'Found {len(replies)} suitable replies, limiting to {self.num_top_replies}...')
            replies = self.quality_score.update_replies(replies, statement)
            replies = sorted(replies, reverse=True)[:self.num_top_replies]

            confidences, texts = list(zip(*replies))
            conf_sums = np.cumsum(confidences)
            roll = np.random.rand() * conf_sums[-1]

            for i, threshold in enumerate(conf_sums):
                if roll < threshold:
                    reply = texts[i]
                    self.log_reply(statement, reply)
                    return reply
            log.error(f"Error rolling dice to select reply. roll: {roll}\nthreshold: {threshold}\nconf_sums: {conf_sums}\n"
                      f"replies: {pd.DataFrame(replies)}")
            return texts[0]
        else:
            log.error(f"No replies ({replies}) were returned by {self.bots}!!")

        reply = "Sorry, something went wrong. Not sure what to say..."
        self.log_reply(statement, reply)
        return reply


def run_skill():
    global BOT
    if BOT is None:
        BOT = CLIBot(
            bots=constants.args.bots,
            num_top_replies=constants.args.num_top_replies,
            semantic=constants.args.semantic,
            sentiment=constants.args.sentiment,
            spell=constants.args.spell)
    if constants.args.persist:
        print('Type "quit" or "exit" to end the conversation...')

    log.debug(f'FINAL PROCESSED ARGS AFTER INSTANTIATING CLIBot:\n{vars(constants.args)}\n')
    return BOT


def cli(args):
    global BOT
    BOT = run_skill() if BOT is None else BOT
    state = {}
    statements = []
    user_statement = ' '.join(args.words)
    statements.append(dict(user=user_statement, bot=None, **state))
    args.persist = args.persist or not len(user_statement)
    for i in range(constants.MAX_TURNS if args.persist else 0):
        if user_statement.lower().strip() in constants.EXIT_COMMANDS:
            break
        if user_statement:
            log.info(f"Computing a reply to {user_statement}...")
            bot_statement = BOT.reply(user_statement)
            statements[-1]['bot'] = bot_statement
            print(f"{args.nickname}: {bot_statement}")
        if args.persist or not user_statement:
            user_statement = input("YOU: ")
            statements.append(dict(user=user_statement, bot=None, **state))
        else:
            break
    return pd.DataFrame(statements)


def main():
    global BOT
    BOT = run_skill() if BOT is None else BOT
    # args = constants.parse_argv(argv=sys.argv)
    statements = cli(constants.args)
    if constants.args.loglevel >= 50:
        return
    return statements


if __name__ == "__main__":
    statements = main()



def score(reply, stmt=None, **kwargs):
    if kwargs is None or kwargs['nlp'] is None:
        return 0.0

    doc = kwargs['nlp'](reply)
    correct_count = sum(
        [(getattr(getattr(token, '_', token), 'hunspell_spell', 0) and not token.is_punct) for token in doc]
    )
    correct_ratio = (correct_count + 1) / (sum([1 if not token.is_punct else 0 for token in doc]) + 1)

    return correct_ratio



""" Download wiktionary slang words glossary """
import os
from datetime import datetime as Datetime
from datetime import date as Date
import requests

import pandas as pd

from qary.constants import DATA_DIR


def scrape_slang(url='https://bestlifeonline.com/2010s-slang/'):
    return requests.get(url).text


def find_latest_index(lang='en', date='20200101'):
    if isinstance(date, (Datetime, Date)):
        date = f'{date.year}{date.month}{date.day}'
    return f'https://dumps.wikimedia.org/{lang}wiktionary/{date}/' \
        f'{lang}wiktionary-{date}-pages-articles-multistream-index.txt.bz2'


def load_english_index(path='{lang}wiktionary-{date}-pages-articles-multistream-index.txt.bz2',
                       lang='en', date='20200101'):
    if '{' in path:
        path = path.format(lang=lang, date=date)
    for p in [path, os.path.join(DATA_DIR, path), find_latest_index]:
        p = p() if callable(p) else p
        try:
            return pd.read_csv(p, sep=':')
        except IOError:
            pass


def get_page_ids(word, url='https://en.wiktionary.org/w/api.php', params=None):
    """ Get the page ID for a word, -1 if page/word doesn't exist

    TODO: search titles with: action=query&list=search&srsearch=Nelson%20Mandela&utf8=&format=json

    >>> get_page_ids('testnotaword')
    []
    """
    default_params = dict(
        action='query',
        format='json',
        list='search',
        # utf8=True,
    )
    default_params.update(dict(title=str(word)))
    default_params.update(params or {})
    resp = requests.get(url, params=default_params)
    js = resp.json()
    if isinstance(js, dict):
        return list(js.get('query', {}).get('pages', {}).keys())
    else:
        return []



def score(reply, stmt=None, **kwargs):
    if kwargs is None or kwargs['sentiment_analyzer'] is None:
        return 0.0

    return 1.0 if kwargs['sentiment_analyzer'].polarity_scores(reply)['compound'] > -0.5 else 0.0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from pathlib import Path
import pytest  # noqa
import yaml
import json

from qary.constants import DATA_DIR

__author__ = "SEE AUTHORS.md"
__copyright__ = "Hobson Lane"
__license__ = "The Hippocratic License, see LICENSE.txt (MIT + Do no Harm)"


def test_yaml_datafiles():
    for path in Path(DATA_DIR).rglob('*.yml'):
        with open(path, 'rt') as fin:
            assert len(yaml.full_load(fin)) >= 0


def test_json_datafiles():
    for path in Path(DATA_DIR).rglob('*.json'):
        with open(path, 'rt') as fin:
            assert len(json.load(fin)) >= 0



# TODO: add doc strings?
import logging
import nltk
import importlib
import sys

from nltk.sentiment.vader import SentimentIntensityAnalyzer

from qary import constants
from qary import spacy_language_model
import qary.scores  # noqa

sys.path.append(constants.BASE_DIR)
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class QualityScore:
    def __init__(self, **kwargs):
        self.metrics = [metric_name for metric_name in kwargs.keys() if kwargs[metric_name]]
        self.weights = list(kwargs.values())
        self.modules = {metric: importlib.import_module(f'qary.scores.{metric}_score') for metric in self.metrics}
        self.nlp = spacy_language_model.nlp
        try:
            self.sentiment_analyzer = SentimentIntensityAnalyzer()
        except LookupError:
            nltk.download('vader_lexicon')
            self.sentiment_analyzer = SentimentIntensityAnalyzer()
        self.kwargs = {'nlp': self.nlp, 'sentiment_analyzer': self.sentiment_analyzer}

    def update_replies(self, replies, stmt=None):
        log.debug(replies)
        metrics_scores = [[reply[0] for reply in replies]]
        for i in range(len(self.metrics)):
            metric = self.metrics[i]
            metrics_scores.append([getattr(self.modules[metric], 'score')(
                reply[1], stmt=stmt, **self.kwargs) for reply in replies])
            metrics_scores[-1] = [float(score + 1) / (max(metrics_scores[-1]) + 1) * self.weights[i]
                                  for score in metrics_scores[-1]]

        updated_replies = list()
        replies_scores = list(map(list, zip(*metrics_scores)))
        for i, reply in enumerate(replies):
            updated_replies.append(((reply[0] * sum(replies_scores[i])) / (len(self.metrics) + 1), reply[1]))

        return updated_replies


def score(reply, stmnt=None):
    """ Combine multiple scores into a single quality score """
    raise NotImplementedError()
